tree_data = [
                        {'label': 'MajorClaim',
                        'link': 1,
                        'link_label': '',
                        'text': 'one who studies overseas will gain many skills throughout this '
                                'experience'},
                        {'label': 'MajorClaim',
                        'link': 1,
                        'link_label': '',
                        'text': 'living and studying overseas gives the individual a new perspective '
                                'on the subject that is studied or in general life'},
                        {'label': 'Claim',
                        'link': 1,
                        'link_label': 'For',
                        'text': 'studying at an overseas university gives individuals the '
                                'opportunity to improve social skills by interacting and '
                                'communicating with students from different origins and cultures'},
                        {'label': 'Claim',
                        'link': 1,
                        'link_label': 'For',
                        'text': 'living and studying overseas is an irreplaceable experience when it '
                                'comes to learn standing on your own feet'},
                        {'label': 'Claim',
                        'link': 1,
                        'link_label': 'For',
                        'text': 'one who has studied and lived overseas will become more eligible '
                                'for the job than his/her peers'},
                        {'label': 'Premise',
                        'link': 2,
                        'link_label': 'Supports',
                        'text': 'Compared to the peers studying in the home country, it will be more '
                                'likely for the one who is living overseas to be successful in '
                                'adapting himself/herself into new environments and situations in '
                                'life'},
                        {'label': 'Premise',
                        'link': 3,
                        'link_label': 'Attacks',
                        'text': 'One who is living overseas will of course struggle with loneliness, '
                                'living away from family and friends'},
                        {'label': 'Premise',
                        'link': 6,
                        'link_label': 'Attacks',
                        'text': 'those difficulties will turn into valuable experiences in the '
                                'following steps of life'},
                        {'label': 'Premise',
                        'link': 3,
                        'link_label': 'Supports',
                        'text': 'the one will learn living without depending on anyone else'},
                        {'label': 'Premise',
                        'link': 4,
                        'link_label': 'Supports',
                        'text': 'employers are mostly looking for people who have international and '
                                'language skills'},
                        {'label': 'Premise',
                        'link': 4,
                        'link_label': 'Supports',
                        'text': 'Becoming successful in this study will give the student an edge in '
                                'job market'},
                        {'label': 'Claim',
                        'link': 1,
                        'link_label': 'Against',
                        'text': 'there are many difficulties a student might face when studying and '
                                'living overseas'}
                        ]





#basics
from copy import deepcopy
import numpy as np

#igraph
import igraph
from igraph import Graph, EdgeSeq

#plotly
import plotly.express as px
import plotly.graph_objects as go


def normalize_row_coordinates(plot_data):
    Y = np.array(plot_data["Y"])
    X = np.array(plot_data["X"], dtype=float)

    # where level is == 1, this is the level of all trees
    # first part only alines the trees so they are equally distributed over the max width
    tree_level_mask = Y==1
    
    nr_trees = len(plot_data["tree_widths"])
    if nr_trees == 1:
        nr_trees += 1

    new_xs = X[tree_level_mask] * ( plot_data["max_width"] /   nr_trees)
    X[tree_level_mask] = new_xs
    plot_data["X"] = X.tolist()

    ## TODO:
    # we need to shift the subtrees of the main tree to make them look a bit nicer
    #
    # possible solutions:
    # shift each level for each tree according to :

    # for tree in trees:
    #     for i in range(2,plot_data["max_depth"]+1):
    #         level_mask = Y == i
    #         tree_maks = root == tree
    #         mask  = level_mask + level_mask
    #         new_xs = X[mask]
    #         new_xs = new_xs * ( plot_data["max_width"]  ( plot_data["tree_widths"][tree] / len(new_xs) )
    #         X[level_mask] = new_xs
    #         plot_data["X"] = X.tolist()


def set_link_coordinates(plot_data, labels:list, colors:list):

    for label in labels:
        plot_data["link_xy"][label] = {
                                        "X":[],
                                        "Y": [],
                                        "style": dict(color=colors.pop(0), width=2, dash='solid')
                                        }
        
    for link_label, idx_of_parent, idx_of_child in plot_data["links"]:
        plot_data["link_xy"][link_label]["X"].extend([plot_data["X"][idx_of_parent], plot_data["X"][idx_of_child], None])
        plot_data["link_xy"][link_label]["Y"].extend([plot_data["Y"][idx_of_parent], plot_data["Y"][idx_of_child], None])


def get_plot_data(data):

    def place_node(tree_nodes, node, plot_data):


        if node["link"] in tree_nodes:
            parent_node = tree_nodes[node["link"]]
            root = parent_node["root"]

            row = parent_node["row"] + 1

            if row not in plot_data["level_widths"]:
                plot_data["level_widths"][row] = 0
            
            plot_data["level_widths"][row] += 1
            column = plot_data["level_widths"][row]

            idx = len(plot_data["X"])
            plot_data["X"].append(column)
            plot_data["Y"].append(row)
            plot_data["root"].append(root)
            plot_data["links"].append((node["link_label"], parent_node["idx"], idx))
            plot_data["labels"].append(node["id"])
            plot_data["max_depth"] = max(plot_data["max_depth"], row)
            plot_data["texts"].append(format_text(node["text"]))
            parent_node["subnodes"][node["id"]] =  {
                                                    "row":row,        
                                                    "column":column,
                                                    "root":root,
                                                    "idx": idx,
                                                    "subnodes":{}
                                                    }
            
            # we need to know the width of a tree so we can normalzie the coordinates later
            plot_data["tree_widths"][root] = max(plot_data["tree_widths"][root], len(parent_node["subnodes"]))
            return True
        else:
            for node_id, node_dict in tree_nodes.items():
                if place_node(node_dict["subnodes"], node, plot_data):
                    return True
                    
            return False

    plot_data = {
                "X":[],
                "Y":[], 
                "links": [],
                "labels": [],
                "texts": [],
                "root":[], 
                "max_depth":0, 
                "tree_widths":{},
                "level_widths":{},
                "link_xy": {}
                }
    trees = {}

    unplaced_nodes = data.copy()
    while unplaced_nodes:

        node  = unplaced_nodes.pop(0)

        #if we have a node that points to itself its the root of a new tree
        if node["id"] == node["link"]:
            row = 1
            column = len(trees)+1
            idx = len(plot_data["X"])
            trees[node["id"]] = {
                            "row":row,        
                            "column":column,
                            "root":node["id"],
                            "idx": idx,
                            "subnodes":{}
                            }
            plot_data["X"].append(column)
            plot_data["Y"].append(row)
            plot_data["root"].append(node["id"])
            plot_data["labels"].append(node["id"])
            plot_data["tree_widths"][node["id"]] = 1
            plot_data["texts"].append(format_text(node["text"]))

            if row not in plot_data["level_widths"]:
                plot_data["level_widths"][row] = 0
        
            plot_data["level_widths"][row] += 1

            found = True
        else:
            found = place_node(trees, node, plot_data)
    
        # if node found a place we remove it else we just add it
        # last to the unplaced_nodes list
        if not found and node not in unplaced_nodes:
            unplaced_nodes.append(node)

    plot_data["max_width"] =  max(plot_data["level_widths"].values())
    return plot_data


def add_node_text(fig, plot_data):
    annotations = [dict(
                                text=l, 
                                x=plot_data["X"][i], 
                                y=plot_data["Y"][i],
                                xref='x1', yref='y1',
                                font=dict(color='black', size=10),
                                showarrow=False
                                ) 
                    for i, l in enumerate(plot_data["labels"])]

    fig.update_layout(
                        annotations=annotations,
                        )


def create_ids(data):
    node_stack = {}
    for node in data:
        label = node["label"]

        if label not in node_stack:
            node_stack[label] = []
        
        node_id = f"{label}_{len(node_stack[label])}"
        
        node_stack[label].append(label)
        node["id"] = node_id


def replacement_legend(fig, legendgroup:str):
    fig.add_trace(go.Scatter(
                            x=[0],
                            y=[0],
                            visible=True,
                            mode="markers",
                            marker=dict(symbol='square',
                                        size=50,
                                        color="white"
                                        ),
                            name=legendgroup,
                            showlegend=True,
                            legendgroup=legendgroup,
                            ))
    fig.update_layout(
                        legend= dict(
                                    font=dict(
                                                family="Courier",
                                                size=15,
                                                color="black"
                                             )
                                    )
                        )
    

def add_lines(fig, plot_data, opacity:float, legendgroup:str):
    for label, link_data in plot_data["link_xy"].items():

        if not label:
            style = dict(color="black", width=2, dash='solid')
            showlegend = False 
        else:
            style = link_data["style"]
            showlegend = True

        fig.add_trace(go.Scatter(
                                x=link_data["X"],
                                y=link_data["Y"],
                                mode='lines',
                                name=label,
                                line=style,
                                hoverinfo="none",
                                showlegend=True if legendgroup is None else False,
                                legendgroup=legendgroup,
                                opacity=opacity,
                                ))


def add_nodes(fig, plot_data, colors:list, opacity:float, legendgroup:str):
    fig.add_trace(go.Scatter(
                            x=plot_data["X"],
                            y=plot_data["Y"],
                            mode='markers+text',
                            name=legendgroup,
                            marker=dict(symbol='diamond-wide',
                                            size=50,
                                            color=colors.pop(0),
                                            #opacity=colors,
                                            ),
                            text=plot_data["labels"],
                            hovertext=plot_data["texts"],
                            hoverinfo='text',
                            showlegend=False,
                            legendgroup=legendgroup,
                            opacity=opacity,
                            textposition='middle center'
                            
                            ))


def format_text(text:str):
    tokens = text.split(" ")
    return "<br>".join([" ".join(tokens[i:i+10]) for i in range(0,len(tokens),10)])


def reformat_links(data):
    label2idx = {d["label"]:i for i,d in enumerate(data)}

    for node in data:
        link  = node["link"]
        if isinstance(link, int):
            node["link_int"] = link
            node["link"] = data[link]["id"]
        else:
            node["link_int"] = label2idx[link]


def create_tree_plot(fig, data:dict, colors:str, reverse:bool, opacity:float=1.0, legendgroup:str=None): # nodes:list, links:list, color:str):

    data = deepcopy(data)
    colors = deepcopy(colors)

    #we turn labels into ids
    if "id" not in data[0]:
        create_ids(data)

    # change links to labels or add int to link_int so we can sort
    reformat_links(data)

    # sort data to untangle trees
    data = sorted(data, key=lambda x:x["link_int"])

    # parse the data to get plot data ( e.g. X Y coordinates, max depth, max width etc)
    plot_data = get_plot_data(data)

    #normalize coordinates to make tree abit nicer now as we know the depth and widths
    normalize_row_coordinates(plot_data)

    # when we have normalized X and Y corridnates we can set up all the links
    link_labels = sorted(set([d["link_label"] for d in data]))
    set_link_coordinates(plot_data, labels=link_labels, colors=colors)
    
    #create the plot
    add_lines(fig, plot_data, opacity=opacity, legendgroup=legendgroup)
    add_nodes(fig, plot_data, colors=colors, opacity=opacity, legendgroup=legendgroup)
    #add_node_text(fig, plot_data)
    
    return plot_data["max_depth"], plot_data["max_width"]


def hot_tree(data, gold_data=None, colors="Plotly", reverse=True, title:str="", save_to:str=None):

    fig = go.Figure()

    #get colors scale
    for scale in [px.colors.sequential, px.colors.qualitative, px.colors.cyclical]:
        if hasattr(scale, colors):
            colors = getattr(scale, colors) 
            break
            
    if not isinstance(colors,list):
        raise KeyError(f"{colors} is not a supported plotly colorscale. scales can be found here: https://plotly.com/python/builtin-colorscales/")
    
    # gold data is added , we can create a tree with high opacity that just sits static in the background
    if gold_data:
        replacement_legend(fig, "gold")
        gold_max_depth, gold_max_width = create_tree_plot(  
                                                            fig, 
                                                            data=gold_data,
                                                            colors=colors,
                                                            reverse=reverse,
                                                            opacity=0.3,
                                                            legendgroup="gold"
                                                            )

    legendgroup = None
    if gold_data:
        replacement_legend(fig, "pred")
        legendgroup = "pred"
    
    max_depth, max_width = create_tree_plot(  
                                            fig, 
                                            data=data,
                                            colors=colors,
                                            reverse=reverse,
                                            legendgroup=legendgroup
                                            )  

    if gold_data:
        max_depth = max(gold_max_depth, max_depth) 
        max_width = max(gold_max_width, max_width)

    fig.update_layout(
                    yaxis=dict(range=[0,max_depth+1], autorange="reversed" if reverse else None),
                    xaxis=dict(range=[0,max_width+1])
                    )
    

    axis = dict(
                showline=False, # hide axis line, grid, ticklabels and  title
                zeroline=False,
                showgrid=False,
                showticklabels=False,
            )

    fig.update_layout(
                        title=title,
                        #annotations=make_annotations(name2pos),
                        font_size=12,
                        showlegend=True,
                        margin=dict(l=40, r=40, b=85, t=100),
                        hovermode='closest',
                        plot_bgcolor="white", #'rgb(248,248,248)',
                        xaxis=axis,
                        yaxis=axis
                        )
    
    if save_to:
        fig.write_image(save_to)

    return fig


    
import numpy as np
import matplotlib.cm as cm
import matplotlib.colors as colors
import json
import imgkit
import base64
from lxml import etree, html
from IPython.display import Image


def class_header(labels:list, style_elems:list):
    header = ["<span> Classes: </span>"] 
    label2cmap = {}
    color_ranges = ["Reds", "Blues", "Greens", "Oranges", "Purples"]
    for i,label in enumerate(labels):
        cmap = cm.get_cmap(color_ranges[i])
        hex_code = colors.to_hex(cmap(1.0))

        style_elems.append(f'.c{label} {{ background-color: {hex_code}; }}')
        header.append(f'<span class="c{label}">  {label}  </span><span> |</span>')
        label2cmap[label] = cmap

    header.append("<br><br>")
    return header, label2cmap

    
def hot_text(data, labels:list, save_path:str="/tmp/hot_text.png", print_html:bool=True, show_spans:bool=True, show_scores:bool=True, font:str="Verdana"):
    
    style_elems = [ 
                    "span { line-height: 54px;}",
                    ".gold {border-top: 3px solid gold; padding-top: 5px;}",
                    ".pred {border-bottom: 3px solid red; padding-bottom: 5px; }",
                    ".gold_pred {border-bottom: 3px solid red; padding-bottom: 5px; border-top: 3px solid gold; padding-top: 5px;}",
                    ".supsub {position: absolute}",
                    ".subscript {display:block; position:relative; left:2px; top: 18px; font-size: 50%; font-weight:bolder; }",
                    ".superscript {display:block; position:relative; left:2px; top:-20px; font-size: 50%; font-weight:bolder;}", #text-shadow: 2px 2px white, 2px -2px white,-2px 2px white,-2px -2px white;
                    ]

    header = []
    if show_scores:
        header, label2cmap = class_header(labels, style_elems)

    gold_spans = []
    pred_spans = []
    token_stack = []

    last_pred_span = ""
    last_gold_span = ""
    for i,td in enumerate(data):

        pred = td["pred"]
        gold = td.get("gold", {"label":None, "span_id":None})
        #token = f'<span>{td["token"]}</span>'
        token = td["token"]

        if show_scores:
            if "score" in pred and "label" in pred:
                if pred["score"] != None and pred["label"] != None:
                    cmap = label2cmap[pred["label"]]
                    hex_code = colors.to_hex(cmap(pred["score"]))
                    style_elems.append(f'.c{i} {{ background-color: {hex_code}; }}')
                    token = f'<span class="c{i}">{token} </span>'

            if ">" not in token:
                token += " "


        
        if show_spans:

            # if we dont have scores we just add some space so that the prediction
            # and gold lines drag over to begining of next word, hence leave the lines intact
            if not show_scores:
                token += " "

            pred_span = pred["span_id"]
            gold_span = gold["span_id"]

            if pred_span and gold_span:
                token = f'<span class="gold_pred">{token}</span>'
            elif pred_span:
                token = f'<span class="pred">{token}</span>'
            elif gold_span:
                token = f'<span class="gold">{token}</span>'

            # we got a new span
            # if gold["span_id"] != last_gold_span and not last_gold_span:
            #     token_stack.append("|")

            # if we have a new span that is not the first for pred
            if last_gold_span != gold["span_id"] and last_gold_span:
                #<gold>|</gold>
                #<span class='supsub'><sup class='superscript'>Sup</sup></span>
                sup = f"<span class='supsub'><sup class='superscript'>{last_gold_span}</sup></span>"
                token_stack.append(sup)

            # if we have a new span that is not the first for gold
            if last_pred_span != pred["span_id"] and last_pred_span:
                #<gold>|</gold>
                sub = f"<span class='supsub'><sub class='subscript'>{last_pred_span}</sub></span>"
                token_stack.append(sub)
            
            last_pred_span = pred["span_id"]    
            last_gold_span = gold["span_id"]
        
        token_stack.append(token)


    #span {{ font-family:"verdana", monospace; font-size: 20px; }} 
    html_string =  f""" <html>
                            <head>
                                <style>
                                {' '.join(style_elems)}
                                </style>
                            </head>
                            <body style="font-family:{font}; font-size:20px;">
                            <br><br>
                            <br><br>
                            {' '.join(header)}
                            {' '.join(token_stack)}

                            <br><br>

                            </body>
                            </html>
                    """

    imgkit.from_string(html_string, save_path, options={'quiet':'', "width": 800, "height":600})


    if print_html:
        document_root = html.fromstring(html_string)
        print(etree.tostring(document_root, encoding='unicode', pretty_print=True))


from setuptools import find_packages, setup
import json
import pathlib
import os


def get_meta():
    meta_file = os.path.join(pathlib.Path(__file__).parent.absolute(), "meta.json")
    with open(meta_file, "r") as f:
        meta = json.load(f)
    return meta

meta = get_meta()


setup(
    name='hotviz',
    version=meta["__version__"],    
    description='A visulizations lib for python',
    url='https://github.com/AxlAlm/hotviz',
    author=meta["__author__"],
    author_email=meta["__author_email__"],
    license=meta["__license__"],
    packages=find_packages(),
    install_requires=[
                        'numpy==1.18.1',
                        'plotly==4.9.0',
                        'python-igraph==0.8.3',
                        'imgkit==1.0.2',                  
                      ],

    classifiers=[
                'Development Status :: 4 - Beta',
                'Programming Language :: Python :: 3.7',
                ],
)
